@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (c) 1998-2005 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (c) 1998-2006 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -56,25 +56,23 @@ static unsigned char *sid_ctx=(unsigned char *)"stunnel SID";
     /* const allowed here */
 #endif
 
-extern SSL_CTX *ctx; /* global SSL context defined in ssl.c */
-
-static int do_client(CLI *);
-static int init_local(CLI *);
-static int init_remote(CLI *);
-static int init_ssl(CLI *);
-static int transfer(CLI *);
-static int parse_socket_error(const char *);
-static void cleanup(CLI *, int);
+static void do_client(CLI *);
+static void run_client(CLI *);
+static void init_local(CLI *);
+static void init_remote(CLI *);
+static void init_ssl(CLI *);
+static void transfer(CLI *);
+static void parse_socket_error(CLI *, const char *);
 
 static void print_cipher(CLI *);
-static int auth_libwrap(CLI *);
-static int auth_user(CLI *);
+static void auth_libwrap(CLI *);
+static void auth_user(CLI *);
 static int connect_local(CLI *);
 #ifndef USE_WIN32
-static int make_sockets(int [2]);
+static void make_sockets(CLI *, int [2]);
 #endif
-static int connect_remote(CLI *c);
-static int connect_wait(int, int);
+static int connect_remote(CLI *);
+static void connect_wait(CLI *);
 static void reset(int, char *);
 
 int max_clients;
@@ -105,64 +103,113 @@ void *client(void *arg) {
 #endif
     s_log(LOG_DEBUG, "%s started", c->opt->servname);
 #ifndef USE_WIN32
-    if(c->opt->option.remote && c->opt->option.program)
-        c->local_rfd.fd=c->local_wfd.fd=connect_local(c);
+    if(c->opt->option.remote && c->opt->option.program) {
             /* connect and exec options specified together */
-            /* spawn local program instead of stdio */
+            /* -> spawn a local program instead of stdio */
+        while((c->local_rfd.fd=c->local_wfd.fd=connect_local(c))>=0) {
+            run_client(c);
+            sleep(1); /* FIXME: not a good idea in ucontext threading */
+        }
+    } else
 #endif
+    {
+        if(alloc_fd(c->local_rfd.fd))
+            return NULL;
+        if(c->local_wfd.fd!=c->local_rfd.fd)
+            if(alloc_fd(c->local_wfd.fd))
+                return NULL;
+        run_client(c);
+    }
+    free(c);
+#ifdef DEBUG_STACK_SIZE
+    stack_info(0); /* display computed value */
+#endif
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
+    _endthread();
+#endif
+#ifdef USE_UCONTEXT
+    s_log(LOG_DEBUG, "Context %ld closed", ready_head->id);
+    s_poll_wait(NULL, 0); /* wait on poll() */
+    s_log(LOG_ERR, "INTERNAL ERROR: failed to drop context");
+#endif
+    return NULL;
+}
+
+static void run_client(CLI *c) {
+    int error;
+
     c->remote_fd.fd=-1;
+    c->fd=-1;
     c->ssl=NULL;
-    cleanup(c, do_client(c));
+    c->sock_bytes=c->ssl_bytes=0;
+
+    error=setjmp(c->err);
+    if(!error)
+        do_client(c);
+
+    s_log(LOG_NOTICE,
+        "Connection %s: %d bytes sent to SSL, %d bytes sent to socket",
+         error ? "reset" : "closed", c->ssl_bytes, c->sock_bytes);
+
+        /* Cleanup IDENT socket */
+    if(c->fd>=0)
+        closesocket(c->fd);
+
+        /* Cleanup SSL */
+    if(c->ssl) { /* SSL initialized */
+        SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+        SSL_free(c->ssl);
+        ERR_remove_state(0);
+    }
+
+        /* Cleanup remote socket */
+    if(c->remote_fd.fd>=0) { /* Remote socket initialized */
+        if(error && c->remote_fd.is_socket)
+            reset(c->remote_fd.fd, "linger (remote)");
+        closesocket(c->remote_fd.fd);
+    }
+
+        /* Cleanup local socket */
+    if(c->local_rfd.fd>=0) { /* Local socket initialized */
+        if(c->local_rfd.fd==c->local_wfd.fd) {
+            if(error && c->local_rfd.is_socket)
+                reset(c->local_rfd.fd, "linger (local)");
+            closesocket(c->local_rfd.fd);
+        } else { /* STDIO */
+            if(error && c->local_rfd.is_socket)
+                reset(c->local_rfd.fd, "linger (local_rfd)");
+            if(error && c->local_wfd.is_socket)
+                reset(c->local_wfd.fd, "linger (local_wfd)");
+       }
+    }
 #ifdef USE_FORK
     if(!c->opt->option.remote) /* 'exec' specified */
-        exec_status(); /* null SIGCHLD handler was used */
+        child_status(); /* null SIGCHLD handler was used */
 #else
     enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
     s_log(LOG_DEBUG, "%s finished (%d left)", c->opt->servname,
         --num_clients);
     leave_critical_section(CRIT_CLIENTS);
 #endif
-    free(c);
-#ifdef DEBUG_STACK_SIZE
-    stack_info(0); /* display computed value */
-#endif
-#ifdef USE_WIN32
-    _endthread();
-#endif
-    return NULL;
 }
 
-static int do_client(CLI *c) {
-    int result;
-
-    if(init_local(c))
-        return -1;
-    if(!options.option.client && !c->opt->protocol) {
+static void do_client(CLI *c) {
+    init_local(c);
+    if(!c->opt->option.client && !c->opt->protocol) {
         /* Server mode and no protocol negotiation needed */
-        if(init_ssl(c))
-            return -1;
-        if(init_remote(c))
-            return -1;
+        init_ssl(c);
+        init_remote(c);
     } else {
-        if(init_remote(c))
-            return -1;
-        if(negotiate(c)<0) {
-            s_log(LOG_ERR, "Protocol negotiations failed");
-            return -1;
-        }
-        if(init_ssl(c))
-            return -1;
+        init_remote(c);
+        negotiate(c);
+        init_ssl(c);
     }
-    result=transfer(c);
-    s_log(LOG_NOTICE,
-        "Connection %s: %d bytes sent to SSL, %d bytes sent to socket",
-         result ? "reset" : "closed", c->ssl_bytes, c->sock_bytes);
-    return result;
+    transfer(c);
 }
 
-static int init_local(CLI *c) {
+static void init_local(CLI *c) {
     SOCKADDR_UNION addr;
-    int addrlen;
+    socklen_t addrlen;
 
     addrlen=sizeof(SOCKADDR_UNION);
     if(getpeername(c->local_rfd.fd, &addr.sa, &addrlen)<0) {
@@ -175,7 +222,7 @@ static int init_local(CLI *c) {
         if(c->opt->option.transparent || get_last_socket_error()!=ENOTSOCK) {
 #endif
             sockerror("getpeerbyname");
-            return -1;
+            longjmp(c->err, 1);
         }
         /* Ignore ENOTSOCK error so 'local' doesn't have to be a socket */
     } else { /* success */
@@ -187,23 +234,15 @@ static int init_local(CLI *c) {
         c->local_wfd.is_socket=1; /* TODO: It's not always true */
         /* It's a socket: lets setup options */
         if(set_socket_options(c->local_rfd.fd, 1)<0)
-            return -1;
-        if(auth_libwrap(c)<0)
-            return -1;
-        if(auth_user(c)<0) {
-            s_log(LOG_WARNING, "Connection from %s REFUSED by IDENT",
-                c->accepting_address);
-            return -1;
-        }
+            longjmp(c->err, 1);
+        auth_libwrap(c);
+        auth_user(c);
         s_log(LOG_NOTICE, "%s connected from %s",
             c->opt->servname, c->accepting_address);
     }
-    return 0; /* OK */
 }
 
-static int init_remote(CLI *c) {
-    int fd;
-
+static void init_remote(CLI *c) {
     /* create connection to host/service */
     if(c->opt->source_addr.num)
         memcpy(&c->bind_addr, &c->opt->source_addr, sizeof(SOCKADDR_LIST));
@@ -214,44 +253,38 @@ static int init_remote(CLI *c) {
     else {
         c->bind_addr.num=0; /* don't bind connecting socket */
     }
-    /* Setup c->remote_fd, now */
+
+    /* setup c->remote_fd, now */
     if(c->opt->option.remote) {
-        fd=connect_remote(c);
+        c->remote_fd.fd=connect_remote(c);
     } else /* NOT in remote mode */
-        fd=connect_local(c);
-    if(fd<0) {
-        s_log(LOG_ERR, "Failed to initialize remote connection");
-        return -1;
-    }
+        c->remote_fd.fd=connect_local(c);
+    c->remote_fd.is_socket=1; /* Always! */
 #ifndef USE_WIN32
-    if(fd>=max_fds) {
+    if(c->remote_fd.fd>=max_fds) {
         s_log(LOG_ERR, "Remote file descriptor out of range (%d>=%d)",
-            fd, max_fds);
-        closesocket(fd);
-        return -1;
+            c->remote_fd.fd, max_fds);
+        longjmp(c->err, 1);
     }
 #endif
-    s_log(LOG_DEBUG, "Remote FD=%d initialized", fd);
-    c->remote_fd.fd=fd;
-    c->remote_fd.is_socket=1; /* Always! */
-    if(set_socket_options(fd, 2)<0)
-        return -1;
-    return 0; /* OK */
+    s_log(LOG_DEBUG, "Remote FD=%d initialized", c->remote_fd.fd);
+    if(set_socket_options(c->remote_fd.fd, 2)<0)
+        longjmp(c->err, 1);
 }
 
-static int init_ssl(CLI *c) {
+static void init_ssl(CLI *c) {
     int i, err;
-    s_poll_set fds;
     SSL_SESSION *old_session;
 
-    if(!(c->ssl=SSL_new(ctx))) {
+    if(!(c->ssl=SSL_new(c->opt->ctx))) {
         sslerror("SSL_new");
-        return -1;
+        longjmp(c->err, 1);
     }
+    SSL_set_ex_data(c->ssl, cli_index, c); /* for verify callback */
 #if SSLEAY_VERSION_NUMBER >= 0x0922
     SSL_set_session_id_context(c->ssl, sid_ctx, strlen(sid_ctx));
 #endif
-    if(options.option.client) {
+    if(c->opt->option.client) {
         if(c->opt->session) {
             enter_critical_section(CRIT_SESSION);
             SSL_set_session(c->ssl, c->opt->session);
@@ -271,7 +304,7 @@ static int init_ssl(CLI *c) {
     }
 
     /* Setup some values for transfer() function */
-    if(options.option.client) {
+    if(c->opt->option.client) {
         c->sock_rfd=&(c->local_rfd);
         c->sock_wfd=&(c->local_wfd);
         c->ssl_rfd=c->ssl_wfd=&(c->remote_fd);
@@ -282,7 +315,7 @@ static int init_ssl(CLI *c) {
     }
 
     while(1) {
-        if(options.option.client)
+        if(c->opt->option.client)
             i=SSL_connect(c->ssl);
         else
             i=SSL_accept(c->ssl);
@@ -290,22 +323,22 @@ static int init_ssl(CLI *c) {
         if(err==SSL_ERROR_NONE)
             break; /* ok -> done */
         if(err==SSL_ERROR_WANT_READ || err==SSL_ERROR_WANT_WRITE) {
-            s_poll_zero(&fds);
-            s_poll_add(&fds, c->ssl_rfd->fd,
+            s_poll_zero(&c->fds);
+            s_poll_add(&c->fds, c->ssl_rfd->fd,
                 err==SSL_ERROR_WANT_READ,
                 err==SSL_ERROR_WANT_WRITE);
-            switch(s_poll_wait(&fds, c->opt->timeout_busy)) {
+            switch(s_poll_wait(&c->fds, c->opt->timeout_busy)) {
             case -1:
                 sockerror("init_ssl: s_poll_wait");
-                return -1; /* error */
+                longjmp(c->err, 1);
             case 0:
                 s_log(LOG_INFO, "init_ssl: s_poll_wait timeout");
-                return -1; /* timeout */
+                longjmp(c->err, 1);
             case 1:
                 break; /* OK */
             default:
                 s_log(LOG_ERR, "init_ssl: s_poll_wait unknown result");
-                return -1; /* error */
+                longjmp(c->err, 1);
             }
             continue; /* ok -> retry */
         }
@@ -316,17 +349,17 @@ static int init_ssl(CLI *c) {
                 continue;
             }
         }
-        if(options.option.client)
+        if(c->opt->option.client)
             sslerror("SSL_connect");
         else
             sslerror("SSL_accept");
-        return -1;
+        longjmp(c->err, 1);
     }
     if(SSL_session_reused(c->ssl)) {
         s_log(LOG_INFO, "SSL %s: previous session reused",
-            options.option.client ? "connected" : "accepted");
+            c->opt->option.client ? "connected" : "accepted");
     } else { /* a new session was negotiated */
-        if(options.option.client) {
+        if(c->opt->option.client) {
             s_log(LOG_INFO, "SSL connected: new session negotiated");
             enter_critical_section(CRIT_SESSION);
             old_session=c->opt->session;
@@ -338,7 +371,6 @@ static int init_ssl(CLI *c) {
             s_log(LOG_INFO, "SSL accepted: new session negotiated");
         print_cipher(c);
     }
-    return 0; /* OK */
 }
 
 /****************************** some defines for transfer() */
@@ -347,20 +379,21 @@ static int init_ssl(CLI *c) {
 #define sock_wr (c->sock_wfd->wr)
 #define ssl_rd  (c->ssl_rfd->rd)
 #define ssl_wr  (c->ssl_wfd->wr)
+
 /* is socket/SSL ready for read/write? */
-#define sock_can_rd (s_poll_canread(&fds, c->sock_rfd->fd))
-#define sock_can_wr (s_poll_canwrite(&fds, c->sock_wfd->fd))
-#define ssl_can_rd  (s_poll_canread(&fds, c->ssl_rfd->fd))
-#define ssl_can_wr  (s_poll_canwrite(&fds, c->ssl_wfd->fd))
+#define sock_can_rd (s_poll_canread(&c->fds, c->sock_rfd->fd))
+#define sock_can_wr (s_poll_canwrite(&c->fds, c->sock_wfd->fd))
+#define ssl_can_rd  (s_poll_canread(&c->fds, c->ssl_rfd->fd))
+#define ssl_can_wr  (s_poll_canwrite(&c->fds, c->ssl_wfd->fd))
 /* NOTE: above defines are related to the logical data stream,
  * no longer to the underlying file descriptors */
+
 /* does the underlaying file descriptor want to read/write? */
 #define want_rd     (SSL_want_read(c->ssl))
 #define want_wr     (SSL_want_write(c->ssl))
 
 /****************************** transfer data */
-static int transfer(CLI *c) {
-    s_poll_set fds;
+static void transfer(CLI *c) {
     int num, err;
     int check_SSL_pending;
     enum {CL_OPEN, CL_INIT, CL_RETRY, CL_CLOSED} ssl_closing=CL_OPEN;
@@ -368,53 +401,52 @@ static int transfer(CLI *c) {
 
     c->sock_ptr=c->ssl_ptr=0;
     sock_rd=sock_wr=ssl_rd=ssl_wr=1;
-    c->sock_bytes=c->ssl_bytes=0;
 
     do { /* main loop */
         /* set flag to try and read any buffered SSL data
          * if we made room in the buffer by writing to the socket */
         check_SSL_pending=0;
 
-        /****************************** setup fds structure */
-        s_poll_zero(&fds); /* Initialize the structure */
+        /****************************** setup c->fds structure */
+        s_poll_zero(&c->fds); /* Initialize the structure */
         if(sock_rd && c->sock_ptr<BUFFSIZE) /* socket input buffer not full*/
-            s_poll_add(&fds, c->sock_rfd->fd, 1, 0);
-        if((ssl_rd && c->ssl_ptr<BUFFSIZE) /* SSL input buffer not full */ ||
+            s_poll_add(&c->fds, c->sock_rfd->fd, 1, 0);
+        if((ssl_rd && c->ssl_ptr<BUFFSIZE) || /* SSL input buffer not full */
                 ((c->sock_ptr || ssl_closing==CL_RETRY) && want_rd))
                 /* want to SSL_write or SSL_shutdown but read from the
                  * underlying socket needed for the SSL protocol */
-            s_poll_add(&fds, c->ssl_rfd->fd, 1, 0);
+            s_poll_add(&c->fds, c->ssl_rfd->fd, 1, 0);
         if(c->ssl_ptr) /* SSL input buffer not empty */
-            s_poll_add(&fds, c->sock_wfd->fd, 0, 1);
-        if(c->sock_ptr /* socket input buffer not empty */ ||
+            s_poll_add(&c->fds, c->sock_wfd->fd, 0, 1);
+        if(c->sock_ptr || /* socket input buffer not empty */
                 ssl_closing==CL_INIT /* need to send close_notify */ ||
                 ((c->ssl_ptr<BUFFSIZE || ssl_closing==CL_RETRY) && want_wr))
                 /* want to SSL_read or SSL_shutdown but write to the
                  * underlying socket needed for the SSL protocol */
-            s_poll_add(&fds, c->ssl_wfd->fd, 0, 1);
+            s_poll_add(&c->fds, c->ssl_wfd->fd, 0, 1);
 
         /****************************** wait for an event */
-        err=s_poll_wait(&fds, (sock_rd && ssl_rd) /* both peers open */ ||
+        err=s_poll_wait(&c->fds, (sock_rd && ssl_rd) /* both peers open */ ||
             c->ssl_ptr /* data buffered to write to socket */ ||
             c->sock_ptr /* data buffered to write to SSL */ ?
             c->opt->timeout_idle : c->opt->timeout_close);
         switch(err) {
         case -1:
-            sockerror("s_poll_wait");
-            return -1;
+            sockerror("transfer: s_poll_wait");
+            longjmp(c->err, 1);
         case 0: /* timeout */
             if((sock_rd && ssl_rd) || c->ssl_ptr || c->sock_ptr) {
                 s_log(LOG_INFO, "s_poll_wait timeout: connection reset");
-                return -1;
+                longjmp(c->err, 1);
             } else { /* already closing connection */
                 s_log(LOG_INFO, "s_poll_wait timeout: connection close");
-                return 0; /* OK */
+                return; /* OK */
             }
         }
         if(!(sock_can_rd || sock_can_wr || ssl_can_rd || ssl_can_wr)) {
             s_log(LOG_ERR, "INTERNAL ERROR: "
                 "s_poll_wait returned %d, but no descriptor is ready", err);
-            return -1;
+            longjmp(c->err, 1);
         }
 
         /****************************** send SSL close_notify message */
@@ -431,7 +463,7 @@ static int transfer(CLI *c) {
                 break;
             case -1: /* a fatal error occurred */
                 sslerror("SSL_shutdown");
-                return -1;
+                longjmp(c->err, 1);
             }
         }
 
@@ -440,8 +472,7 @@ static int transfer(CLI *c) {
             num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
             switch(num) {
             case -1: /* error */
-                if(parse_socket_error("writesocket"))
-                    return -1;
+                parse_socket_error(c, "writesocket");
                 break;
             case 0:
                 s_log(LOG_DEBUG, "No data written to the socket: retrying");
@@ -457,10 +488,8 @@ static int transfer(CLI *c) {
         }
 
         /****************************** write to SSL */
-        if(ssl_wr && ( /* SSL is open for write */
-                (c->sock_ptr && ssl_can_wr) ||
-                /* application data can be written */
-                (want_rd && ssl_can_rd)
+        if(ssl_wr && c->sock_ptr && ( /* output buffer not empty */
+                ssl_can_wr || (want_rd && ssl_can_rd)
                 /* SSL_write wants to read from the underlying descriptor */
                 )) {
             num=SSL_write(c->ssl, c->sock_buff, c->sock_ptr);
@@ -482,8 +511,8 @@ static int transfer(CLI *c) {
                     "SSL_write returned WANT_X509_LOOKUP: retrying");
                 break;
             case SSL_ERROR_SYSCALL: /* really an error */
-                if(num && parse_socket_error("SSL_write"))
-                    return -1;
+                if(num)
+                    parse_socket_error(c, "SSL_write");
                 break;
             case SSL_ERROR_ZERO_RETURN: /* close_notify received */
                 s_log(LOG_DEBUG, "SSL closed on SSL_write");
@@ -491,10 +520,10 @@ static int transfer(CLI *c) {
                 break;
             case SSL_ERROR_SSL:
                 sslerror("SSL_write");
-                return -1;
+                longjmp(c->err, 1);
             default:
                 s_log(LOG_ERR, "SSL_write/SSL_get_error returned %d", err);
-                return -1;
+                longjmp(c->err, 1);
             }
         }
 
@@ -504,8 +533,7 @@ static int transfer(CLI *c) {
                 c->sock_buff+c->sock_ptr, BUFFSIZE-c->sock_ptr);
             switch(num) {
             case -1:
-                if(parse_socket_error("readsocket"))
-                    return -1;
+                parse_socket_error(c, "readsocket");
                 break;
             case 0: /* close */
                 s_log(LOG_DEBUG, "Socket closed on read");
@@ -518,10 +546,8 @@ static int transfer(CLI *c) {
         }
 
         /****************************** read from SSL */
-        if(ssl_rd && ( /* SSL is open for read */
-                (c->ssl_ptr<BUFFSIZE && ssl_can_rd) ||
-                /* there's any application data coming in */
-                (want_wr && ssl_can_wr) ||
+        if(ssl_rd && c->ssl_ptr<BUFFSIZE  && ( /* input buffer not full */
+                ssl_can_rd || (want_wr && ssl_can_wr) ||
                 /* SSL_read wants to write to the underlying descriptor */
                 (check_SSL_pending && SSL_pending(c->ssl))
                 /* write made space from full buffer */
@@ -548,13 +574,13 @@ static int transfer(CLI *c) {
                         s_log(LOG_ERR,
                             "SSL socket closed with %d byte(s) in buffer",
                             c->sock_ptr);
-                        return -1; /* reset the socket */
+                        longjmp(c->err, 1); /* reset the socket */
                     }
                     s_log(LOG_DEBUG, "SSL socket closed on SSL_read");
                     ssl_rd=ssl_wr=0; /* buggy or SSLv2 peer: no close_notify */
                     ssl_closing=CL_CLOSED; /* don't try to send it back */
-                } else if(parse_socket_error("SSL_read"))
-                    return -1;
+                } else
+                    parse_socket_error(c, "SSL_read");
                 break;
             case SSL_ERROR_ZERO_RETURN: /* close_notify received */
                 s_log(LOG_DEBUG, "SSL closed on SSL_read");
@@ -562,10 +588,10 @@ static int transfer(CLI *c) {
                 break;
             case SSL_ERROR_SSL:
                 sslerror("SSL_read");
-                return -1;
+                longjmp(c->err, 1);
             default:
                 s_log(LOG_ERR, "SSL_read/SSL_get_error returned %d", err);
-                return -1;
+                longjmp(c->err, 1);
             }
         }
 
@@ -589,9 +615,19 @@ static int transfer(CLI *c) {
                 ssl_closing=CL_CLOSED; /* closed */
             }
         }
+        if(ssl_closing==CL_RETRY) { /* SSL shutdown */
+            if(!want_rd && !want_wr) { /* close_notify alert was received */
+                s_log(LOG_DEBUG, "SSL doesn't need to read or write");
+                ssl_closing=CL_CLOSED;
+            }
+            if(watchdog>5) {
+                s_log(LOG_NOTICE, "Too many retries on SSL shutdown");
+                ssl_closing=CL_CLOSED;
+            }
+        }
 
         /****************************** check watchdog */
-        if(++watchdog>1000) { /* loop executes without transferring any data */
+        if(++watchdog>100) { /* loop executes without transferring any data */
             s_log(LOG_ERR,
                 "transfer() loop executes not transferring any data");
             s_log(LOG_ERR,
@@ -604,62 +640,33 @@ static int transfer(CLI *c) {
                 ssl_can_rd ? "yes" : "no", ssl_can_wr ? "yes" : "no");
             s_log(LOG_ERR, "ssl want: rd=%s wr=%s",
                 want_rd ? "yes" : "no", want_wr ? "yes" : "no");
-            s_log(LOG_ERR, "socket write buffer: %d byte(s), "
-                "ssl write buffer: %d byte(s)", c->sock_ptr, c->ssl_ptr);
+            s_log(LOG_ERR, "socket input buffer: %d byte(s), "
+                "ssl input buffer: %d byte(s)", c->sock_ptr, c->ssl_ptr);
             s_log(LOG_ERR, "check_SSL_pending=%d, ssl_closing=%d",
                 check_SSL_pending, ssl_closing);
-            return -1;
+            longjmp(c->err, 1);
         }
 
     } while(sock_wr || ssl_closing!=CL_CLOSED);
-
-    return 0; /* OK */
 }
 
-static int parse_socket_error(const char *text) {
+static void parse_socket_error(CLI *c, const char *text) {
     switch(get_last_socket_error()) {
     case EINTR:
         s_log(LOG_DEBUG, "%s interrupted by a signal: retrying", text);
-        return 0;
+        return;
     case EWOULDBLOCK:
         s_log(LOG_NOTICE, "%s would block: retrying", text);
-        return 0;
+        sleep(1); /* Microsoft bug KB177346 */
+        return;
 #if EAGAIN!=EWOULDBLOCK
     case EAGAIN:
         s_log(LOG_DEBUG, "%s temporary lack of resources: retrying", text);
-        return 0;
+        return;
 #endif
     default:
         sockerror(text);
-        return -1;
-    }
-}
-
-static void cleanup(CLI *c, int error) {
-        /* Cleanup SSL */
-    if(c->ssl) { /* SSL initialized */
-        SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-        SSL_free(c->ssl);
-        ERR_remove_state(0);
-    }
-        /* Cleanup remote socket */
-    if(c->remote_fd.fd>=0) { /* Remote socket initialized */
-        if(error && c->remote_fd.is_socket)
-            reset(c->remote_fd.fd, "linger (remote)");
-        closesocket(c->remote_fd.fd);
-    }
-        /* Cleanup local socket */
-    if(c->local_rfd.fd>=0) { /* Local socket initialized */
-        if(c->local_rfd.fd==c->local_wfd.fd) {
-            if(error && c->local_rfd.is_socket)
-                reset(c->local_rfd.fd, "linger (local)");
-            closesocket(c->local_rfd.fd);
-        } else { /* STDIO */
-            if(error && c->local_rfd.is_socket)
-                reset(c->local_rfd.fd, "linger (local_rfd)");
-            if(error && c->local_wfd.is_socket)
-                reset(c->local_wfd.fd, "linger (local_wfd)");
-       }
+        longjmp(c->err, 1);
     }
 }
 
@@ -681,88 +688,115 @@ static void print_cipher(CLI *c) { /* print negotiated cipher */
 #endif
 }
 
-static int auth_libwrap(CLI *c) {
+static void auth_libwrap(CLI *c) {
 #ifdef USE_LIBWRAP
     struct request_info request;
-    int result;
+    int fd[2];
+    int result=0; /* deny by default */
 
-    enter_critical_section(CRIT_INET); /* libwrap is not mt-safe */
-    request_init(&request,
-        RQ_DAEMON, c->opt->servname, RQ_FILE, c->local_rfd.fd, 0);
-    fromhost(&request);
-    result=hosts_access(&request);
-    leave_critical_section(CRIT_INET);
+    if(pipe(fd)<0) {
+        ioerror("pipe");
+        longjmp(c->err, 1);
+    }
+    if(alloc_fd(fd[0]) || alloc_fd(fd[1]))
+        longjmp(c->err, 1);
+    switch(fork()) {
+    case -1:    /* error */
+        close(fd[0]);
+        close(fd[1]);
+        ioerror("fork");
+        longjmp(c->err, 1);
+    case  0:    /* child */
+        close(fd[0]); /* read side */
+        request_init(&request,
+            RQ_DAEMON, c->opt->servname, RQ_FILE, c->local_rfd.fd, 0);
+        fromhost(&request);
+        result=hosts_access(&request);
+        write_blocking(c, fd[1], (u8 *)&result, sizeof(result));
+            /* ignore the returned error */
+        close(fd[1]); /* write side */
+        _exit(0);
+    default:    /* parent */
+        close(fd[1]); /* write side */
+        read_blocking(c, fd[0], (u8 *)&result, sizeof(result));
+            /* ignore the returned error */
+        close(fd[0]); /* read side */
+        /* no need to wait() for zombies here:
+         *  - in UCONTEXT/PTHREAD mode they're removed using the signal pipe
+         *  - in FORK mode they're removed with the client process */
+    }
 
-    if (!result) {
+    if(!result) {
         s_log(LOG_WARNING, "Connection from %s REFUSED by libwrap",
             c->accepting_address);
-        s_log(LOG_DEBUG, "See hosts_access(5) for details");
-        return -1; /* FAILED */
+        s_log(LOG_DEBUG, "See hosts_access(5) manual for details");
+        longjmp(c->err, 1);
     }
+    s_log(LOG_DEBUG, "Connection from %s permitted by libwrap",
+        c->accepting_address);
 #endif
-    return 0; /* OK */
 }
 
-static int auth_user(CLI *c) {
+static void auth_user(CLI *c) {
+#ifndef _WIN32_WCE
     struct servent *s_ent;    /* structure for getservbyname */
+#endif
     SOCKADDR_UNION ident;     /* IDENT socket name */
-    int fd;                   /* IDENT socket descriptor */
     char name[STRLEN];
-    int retval;
     int error;
 
     if(!c->opt->username)
-        return 0; /* -u option not specified */
-    if((fd=socket(c->peer_addr.addr[0].sa.sa_family, SOCK_STREAM, 0))<0) {
+        return; /* -u option not specified */
+    if((c->fd=
+            socket(c->peer_addr.addr[0].sa.sa_family, SOCK_STREAM, 0))<0) {
         sockerror("socket (auth_user)");
-        return -1;
+        longjmp(c->err, 1);
     }
-    if(alloc_fd(fd))
-        return -1;
+    if(alloc_fd(c->fd))
+        longjmp(c->err, 1);
     memcpy(&ident, &c->peer_addr.addr[0], sizeof(SOCKADDR_UNION));
+#ifndef _WIN32_WCE
     s_ent=getservbyname("auth", "tcp");
     if(s_ent) {
         ident.in.sin_port=s_ent->s_port;
-    } else {
+    } else
+#endif
+    {
         s_log(LOG_WARNING, "Unknown service 'auth': using default 113");
         ident.in.sin_port=htons(113);
     }
-    if(connect(fd, &ident.sa, addr_len(ident))) {
+    if(connect(c->fd, &ident.sa, addr_len(ident))) {
         error=get_last_socket_error();
         if(error!=EINPROGRESS && error!=EWOULDBLOCK) {
             sockerror("ident connect (auth_user)");
-            closesocket(fd);
-            return -1;
-        }
-        if(connect_wait(fd, c->opt->timeout_connect)) { /* error */
-            closesocket(fd);
-            return -1;
+            longjmp(c->err, 1);
         }
+        connect_wait(c);
     }
     s_log(LOG_DEBUG, "IDENT server connected");
-    if(fdprintf(c, fd, "%u , %u",
-            ntohs(c->peer_addr.addr[0].in.sin_port),
-            ntohs(c->opt->local_addr.addr[0].in.sin_port))<0) {
-        sockerror("fdprintf (auth_user)");
-        closesocket(fd);
-        return -1;
-    }
-    if(fdscanf(c, fd, "%*[^:]: USERID :%*[^:]:%s", name)!=1) {
+    fdprintf(c, c->fd, "%u , %u",
+        ntohs(c->peer_addr.addr[0].in.sin_port),
+        ntohs(c->opt->local_addr.addr[0].in.sin_port));
+    if(fdscanf(c, c->fd, "%*[^:]: USERID :%*[^:]:%s", name)!=1) {
         s_log(LOG_ERR, "Incorrect data from IDENT server");
-        closesocket(fd);
-        return -1;
+        longjmp(c->err, 1);
     }
-    closesocket(fd);
-    retval=strcmp(name, c->opt->username) ? -1 : 0;
-    safestring(name);
-    s_log(LOG_INFO, "IDENT resolved remote user to %s", name);
-    return retval;
+    closesocket(c->fd);
+    c->fd=-1; /* avoid double close at cleanup */
+    if(strcmp(name, c->opt->username)) {
+        safestring(name);
+        s_log(LOG_WARNING, "Connection from %s REFUSED by IDENT (user %s)",
+            c->accepting_address, name);
+        longjmp(c->err, 1);
+    }
+    s_log(LOG_INFO, "IDENT authentication passed");
 }
 
 static int connect_local(CLI *c) { /* spawn local process */
 #if defined (USE_WIN32) || defined (__vms)
     s_log(LOG_ERR, "LOCAL MODE NOT SUPPORTED ON WIN32 and OpenVMS PLATFORM");
-    return -1;
+    longjmp(c->err, 1);
+    return -1; /* some C compilers require a return value */
 #else /* USE_WIN32, __vms */
     char env[3][STRLEN], name[STRLEN], *portname;
     int fd[2], pid;
@@ -774,14 +808,11 @@ static int connect_local(CLI *c) { /* spawn local process */
     if (c->opt->option.pty) {
         char tty[STRLEN];
 
-        if(pty_allocate(fd, fd+1, tty, STRLEN)) {
-            return -1;
-        }
+        if(pty_allocate(fd, fd+1, tty, STRLEN))
+            longjmp(c->err, 1);
         s_log(LOG_DEBUG, "%s allocated", tty);
-    } else {
-        if(make_sockets(fd))
-            return -1;
-    }
+    } else
+        make_sockets(c, fd);
     pid=fork();
     c->pid=(unsigned long)pid;
     switch(pid) {
@@ -789,7 +820,7 @@ static int connect_local(CLI *c) { /* spawn local process */
         closesocket(fd[0]);
         closesocket(fd[1]);
         ioerror("fork");
-        return -1;
+        longjmp(c->err, 1);
     case  0:    /* child */
         closesocket(fd[0]);
         dup2(fd[1], 0);
@@ -846,19 +877,19 @@ static int connect_local(CLI *c) { /* spawn local process */
 
 #ifndef USE_WIN32
 
-static int make_sockets(int fd[2]) { /* make a pair of connected sockets */
+static void make_sockets(CLI *c, int fd[2]) { /* make a pair of connected sockets */
 #ifdef INET_SOCKET_PAIR
     SOCKADDR_UNION addr;
-    int addrlen;
+    socklen_t addrlen;
     int s; /* temporary socket awaiting for connection */
 
     if((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
         sockerror("socket#1");
-        return -1;
+        longjmp(c->err, 1);
     }
     if((fd[1]=socket(AF_INET, SOCK_STREAM, 0))<0) {
         sockerror("socket#2");
-        return -1;
+        longjmp(c->err, 1);
     }
     addrlen=sizeof(SOCKADDR_UNION);
     memset(&addr, 0, addrlen);
@@ -871,44 +902,44 @@ static int make_sockets(int fd[2]) { /* make a pair of connected sockets */
         log_error(LOG_DEBUG, get_last_socket_error(), "bind#2");
     if(listen(s, 5)) {
         sockerror("listen");
-        return -1;
+        longjmp(c->err, 1);
     }
     if(getsockname(s, &addr.sa, &addrlen)) {
         sockerror("getsockname");
-        return -1;
+        longjmp(c->err, 1);
     }
     if(connect(fd[1], &addr.sa, addrlen)) {
         sockerror("connect");
-        return -1;
+        longjmp(c->err, 1);
     }
     if((fd[0]=accept(s, &addr.sa, &addrlen))<0) {
         sockerror("accept");
-        return -1;
+        longjmp(c->err, 1);
     }
     closesocket(s); /* don't care about the result */
 #else
     if(socketpair(AF_UNIX, SOCK_STREAM, 0, fd)) {
         sockerror("socketpair");
-        return -1;
+        longjmp(c->err, 1);
     }
 #endif
-    return 0;
 }
 #endif
 
 static int connect_remote(CLI *c) { /* connect to remote host */
     SOCKADDR_UNION bind_addr, addr;
     SOCKADDR_LIST resolved_list, *address_list;
-    int error;
-    int s; /* destination socket */
+    int error, fd;
     u16 i;
 
     /* setup address_list */
     if(c->opt->option.delayed_lookup) {
         resolved_list.num=0;
         if(!name2addrlist(&resolved_list,
-                c->opt->remote_address, DEFAULT_LOOPBACK))
-            return -1; /* no host resolved */
+                c->opt->remote_address, DEFAULT_LOOPBACK)){
+            s_log(LOG_ERR, "No host resolved");
+            longjmp(c->err, 1);
+        }
         address_list=&resolved_list;
     } else /* use pre-resolved addresses */
         address_list=&c->opt->remote_addr;
@@ -920,19 +951,18 @@ static int connect_remote(CLI *c) { /* connect to remote host */
         address_list->cur=(address_list->cur+1)%address_list->num;
         /* race condition is possible, but harmless in this case */
 
-        if((s=socket(addr.sa.sa_family, SOCK_STREAM, 0))<0) {
+        if((c->fd=socket(addr.sa.sa_family, SOCK_STREAM, 0))<0) {
             sockerror("remote socket");
-            return -1;
+            longjmp(c->err, 1);
         }
-        if(alloc_fd(s))
-            return -1;
+        if(alloc_fd(c->fd))
+            longjmp(c->err, 1);
 
         if(c->bind_addr.num) { /* explicit local bind or transparent proxy */
             memcpy(&bind_addr, &c->bind_addr.addr[0], sizeof(SOCKADDR_UNION));
-            if(bind(s, &bind_addr.sa, addr_len(bind_addr))<0) {
+            if(bind(c->fd, &bind_addr.sa, addr_len(bind_addr))<0) {
                 sockerror("bind transparent");
-                closesocket(s);
-                return -1;
+                longjmp(c->err, 1);
             }
         }
 
@@ -940,57 +970,66 @@ static int connect_remote(CLI *c) { /* connect to remote host */
         s_ntop(c->connecting_address, &addr);
         s_log(LOG_DEBUG, "%s connecting %s",
             c->opt->servname, c->connecting_address);
-        if(!connect(s, &addr.sa, addr_len(addr)))
-            return s; /* no error -> success (should not be possible) */
+        if(!connect(c->fd, &addr.sa, addr_len(addr))) {
+            fd=c->fd;
+            c->fd=-1;
+            return fd; /* no error -> success (should not be possible) */
+        }
         error=get_last_socket_error();
         if(error!=EINPROGRESS && error!=EWOULDBLOCK) {
             s_log(LOG_ERR, "remote connect (%s): %s (%d)",
                 c->connecting_address, my_strerror(error), error);
-            closesocket(s);
+            closesocket(c->fd);
+            c->fd=-1;
             continue; /* next IP */
         }
-        if(!connect_wait(s, c->opt->timeout_connect))
-            return s; /* success! */
-        closesocket(s); /* error -> next IP */
+        connect_wait(c);
+        fd=c->fd;
+        c->fd=-1;
+        return fd; /* success! */
     }
-    return -1;
+    longjmp(c->err, 1);
+    return -1; /* some C compilers require a return value */
 }
 
     /* wait for the result of a non-blocking connect */
-static int connect_wait(int fd, int timeout) {
-    s_poll_set fds;
-    int error, len;
+    /* file descriptor : c->fd                       */
+    /* timeout         : c->opt->timeout_connect     */
+static void connect_wait(CLI *c) {
+    int error;
+    socklen_t optlen;
 
-    s_log(LOG_DEBUG, "connect_wait: waiting %d seconds", timeout);
-    s_poll_zero(&fds);
-    s_poll_add(&fds, fd, 1, 1);
-    switch(s_poll_wait(&fds, timeout)) {
+    s_log(LOG_DEBUG, "connect_wait: waiting %d seconds",
+        c->opt->timeout_connect);
+    s_poll_zero(&c->fds);
+    s_poll_add(&c->fds, c->fd, 1, 1);
+    switch(s_poll_wait(&c->fds, c->opt->timeout_connect)) {
     case -1:
         sockerror("connect_wait: s_poll_wait");
-        return -1; /* error */
+        longjmp(c->err, 1);
     case 0:
         s_log(LOG_INFO, "connect_wait: s_poll_wait timeout");
-        return -1; /* error */
+        longjmp(c->err, 1);
     default:
-        if(s_poll_canread(&fds, fd)) {
+        if(s_poll_canread(&c->fds, c->fd)) {
             /* just connected socket should not be ready for read */
             /* get the resulting error code, now */
-            len=sizeof(error);
-            if(!getsockopt(fd, SOL_SOCKET, SO_ERROR, (void *)&error, &len))
+            optlen=sizeof(error);
+            if(!getsockopt(c->fd, SOL_SOCKET, SO_ERROR, (void *)&error, &optlen))
                 errno=error;
             if(errno) { /* really an error? */
                 sockerror("connect_wait: getsockopt");
-                return -1; /* connection failed */
+                longjmp(c->err, 1);
             }
         }
-        if(s_poll_canwrite(&fds, fd)) {
+        if(s_poll_canwrite(&c->fds, c->fd)) {
             s_log(LOG_DEBUG, "connect_wait: connected");
-            return 0; /* success */
+            return; /* success */
         }
         s_log(LOG_ERR, "connect_wait: unexpected s_poll_wait result");
-        return -1; /* unexpected result */
+        longjmp(c->err, 1);
     }
-    return -1; /* should not be possible */
+    longjmp(c->err, 1); /* should not be possible */
 }
 
 static void reset(int fd, char *txt) {
